#[cfg(test)]
mod tests;

use std::iter::Peekable;
use std::vec::IntoIter;

#[derive(Debug, PartialEq)]
pub enum Token {
    LeftParen,
    RightParen,
    Symbol(String),
}

#[derive(Debug, PartialEq, Clone)]
pub enum Ast {
    Symbol(String),
    List(Vec<Ast>),
    Root(Vec<Ast>),
}

pub fn tokenize(input: &str) -> Vec<Token> {
    let mut tokens = Vec::new();
    let mut chars = input.chars().peekable();
    while let Some(c) = chars.next() {
        match c {
            '(' => tokens.push(Token::LeftParen),
            ')' => tokens.push(Token::RightParen),
            _ if c.is_whitespace() => (),
            _ => {
                let mut symbol = c.to_string();
                while let Some(&c) = chars.peek() {
                    if c.is_whitespace() || c == '(' || c == ')' {
                        break;
                    }
                    symbol.push(c);
                    chars.next();
                }
                tokens.push(Token::Symbol(symbol));
            }
        }
    }
    tokens
}

fn parse_expr(tokens: &mut Peekable<IntoIter<Token>>) -> Ast {
    match tokens.next() {
        Some(Token::LeftParen) => {
            let mut list = Vec::new();
            while tokens.peek() != Some(&Token::RightParen) {
                list.push(parse_expr(tokens));
            }
            tokens.next();
            Ast::List(list)
        }
        Some(Token::RightParen) => panic!("unexpected )"),
        Some(Token::Symbol(s)) => Ast::Symbol(s),
        None => panic!("unexpected EOF"),
    }
}

pub fn parse(tokens: Vec<Token>) -> Ast {
    let mut tokens = tokens.into_iter().peekable();
    let mut ast: Vec<Ast> = Vec::new();
    while tokens.peek().is_some() {
        ast.push(parse_expr(&mut tokens));
    }
    Ast::Root(ast)
}